-- autostack
-- automatically go through the motions of stacking avalons

local tick = Timer()
local state = "stop"

local SLOT = 2
local SHOOTCMD = "+shoot"..tostring(SLOT)
local STOPSHOOTCMD = "+shoot"..tostring(SLOT).." 0"

local mode

local launchtime = 0
local targetdist = 0
local weaponspeed = 35
local launchspeed = 0
local ammo = 0
local maxspeed
local thrust = 0

local function asprint(str)
	print("autostack: "..tostring(str))
end

local function difftime(t2, t1)
	return (t2 - t1) / 1000
end

local function get_weapon_id(slot)
	local ids = {GetActiveShipSelectedWeaponGroupIDs()}
	local id
	if slot == 1 then
		id = ids[2]
	elseif slot == 2 then
		id = ids[1]
	else
		id = ids[3]
	end
	local port = next(GetActiveShipWeaponGroup(id))
	return GetActiveShipItemIDAtPort(port)
end

local function get_ammo_count(slot)
	local weapon = get_weapon_id(slot)
	if not weapon then
		return
	end

	return GetAddonItemInfo(get_weapon_id(slot))
end

local function get_weapon_speed(slot)
	local weapon = get_weapon_id(slot)
	if not weapon then 
		return
	end

	return tonumber(GetInventoryItemExtendedDesc(weapon):match(".*Velocity: (%d+)m/s.*"))
end

local function ready_to_brake()
	local shipspeed = GetActiveShipSpeed()
	if maxspeed then
		if shipspeed > maxspeed then
			return true
		end
		return
	end
	
	if shipspeed < weaponspeed then
		return
	end

	local shipmaxspeed = GetActiveShipMaxSpeed()
	local mass = GetActiveShipMass()

	local deaccel = thrust / mass
	local turbodeaccel = deaccel * 2
	

	local deacceltime = 0
	local sdisp = 0
	if shipspeed > shipmaxspeed then
		-- above turbo deacceal is x2 regular
		local turbodeacceltime = (shipspeed - shipmaxspeed) / (deaccel * 2)
		sdisp = ((shipspeed + shipmaxspeed) / 2) * turbodeacceltime

		-- below turbo speed it seems to approach the regular deaccel.. x1.5 instead should be close enough
		local noturbodeacceltime = (shipmaxspeed - launchspeed) / (deaccel * 1.5)
		sdisp = sdisp + (((shipspeed + launchspeed) / 2) * noturbodeacceltime)

		deacceltime = turbodeacceltime + noturbodeacceltime
	else
		-- use just the regular deaccl in this case
		deacceltime = (shipspeed - launchspeed) / deaccel
		sdisp = ((shipspeed + launchspeed) / 2) * deacceltime		
	end

	-- compensate for wobble when target speed is reached.. totally made up number
	deacceltime = deacceltime + (50 / deaccel)

	local sdist = GetTargetDistance() or math.huge
	local expectedsdist = sdist - sdisp


	local wtraveltime = difftime(gkmisc.GetGameTime(), launchtime)
	local wdist = targetdist - (wtraveltime * weaponspeed)
	local wdisp = deacceltime * weaponspeed
	local expectedwtraveldist = wdisp + (wtraveltime * weaponspeed)
	local expectedwdist = targetdist - expectedwtraveldist
	

	if sdist < wdist and expectedsdist < expectedwdist then
		return true
	end
end


local function checkstate()
	-- XXX race potential
	if PlayerInStation() then 
		asprint("you are docked!")
		state = "stop"
	end

	if state == "start" then
		if not GetActiveShipID() then
			asprint("you don't have a ship!")
			state = "stop"
			tick:SetTimeout(100, checkstate)
			return
		end

		if not GetTargetDistance() then
			asprint("you don't have a anything targeted!")
			state = "stop"
			tick:SetTimeout(100, checkstate)
			return
		end


		Game.SetCVar("flymodeflag", 1)
		Game.SetDesiredSpeed(0)
		gkinterface.GKProcessCommand("+turbo 0")

		if math.floor(GetActiveShipSpeed()) == 0 then
			thrust = tonumber(GetInventoryItemExtendedDesc(GetActiveShipID()):match("Thrust: (%d+) N.*"))

			weaponspeed = get_weapon_speed(SLOT)
			if not weaponspeed then
				asprint("can't determine weapon speed. check your weapon groups")
				state = "stop"
				tick:SetTimeout(100, checkstate)
				return
			end
			weaponspeed = weaponspeed + launchspeed
			
			if maxspeed and weaponspeed > maxspeed then
				asprint("warning max speed too low")
			end

			if launchspeed > GetActiveShipMaxSpeed() then
				asprint(string.format("warning launch speed (%im/s) should be lower than the max non turbo speed (%im/s)", launchspeed, GetActiveShipMaxSpeed()))				
			end

			if launchspeed == 0 then
				state = "start_shoot"
			else
				Game.SetDesiredSpeed(launchspeed)
				-- pretend that we fired so that we have something to chase for the initial acceleration cycle
				launchtime = gkmisc.GetGameTime()
				targetdist = GetTargetDistance()
				state = "accel"
			end
		end
		
		tick:SetTimeout(100, checkstate)
	elseif state == "start_shoot" then
		ammo = get_ammo_count(SLOT)
		if not ammo then
			asprint("not an ammo based weapon. check your weapon groups")
			state = "stop"
			tick:SetTimeout(100, checkstate)
			return
		end

		gkinterface.GKProcessCommand(SHOOTCMD)
		launchtime = gkmisc.GetGameTime()
		targetdist = GetTargetDistance() or math.huge

		state = "waitforshoot"
		tick:SetTimeout(100, checkstate)
	elseif state == "brake" then
		gkinterface.GKProcessCommand("+turbo 0")
		if math.floor(GetActiveShipSpeed()) <= launchspeed then
			state = "waitfordist"
		end
		tick:SetTimeout(100, checkstate)
	elseif state == "waitfordist" then
		local traveltime = difftime(gkmisc.GetGameTime(), launchtime)
		local traveldist = traveltime * weaponspeed
		local curtargetdist = GetTargetDistance() or math.huge
		if (targetdist - traveldist) < curtargetdist then
			state = "shoot"
		end
		tick:SetTimeout(100, checkstate)
	elseif state == "shoot" then
		ammo = get_ammo_count(SLOT)

		gkinterface.GKProcessCommand(SHOOTCMD)
		state = "waitforshoot"
		tick:SetTimeout(100, checkstate)
	elseif state == "waitforshoot" then
		local tmpammo = get_ammo_count(SLOT)
		if tmpammo == 0 then
			state = "stop"
		elseif tmpammo < ammo then
			state = "stopshoot"
		end
		tick:SetTimeout(100, checkstate)
	elseif state == "stopshoot" then
		gkinterface.GKProcessCommand(STOPSHOOTCMD)
		state = "accel"
		tick:SetTimeout(100, checkstate)
	elseif state == "accel" then
		gkinterface.GKProcessCommand("+turbo")
		if ready_to_brake() then
			state = "brake"
		end
		tick:SetTimeout(100, checkstate)
	elseif state == "stop" then
		Game.SetDesiredSpeed(0)
		gkinterface.GKProcessCommand("+turbo 0")
		gkinterface.GKProcessCommand(STOPSHOOTCMD)		
		asprint("done!")
	else
		asprint("unknown state "..tostring(state))
		state = "stop"
	end
end


local function start()
	state = "start"
	checkstate()
end

local function stop()
	tick:Kill()
	state = "stop"
	checkstate()
end

local function slave_start()
	SendChat(string.format("autostack start %s %s", launchspeed or "", maxspeed or ""), "GROUP")
end

local autostack = {}
function autostack:OnEvent(ev, data)
	if not mode then
		return
	end

	if mode == "master" then
		if ev == "INVENTORY_UPDATE" and GetInventoryItemClassType(data) == CLASSTYPE_ADDON and GetTargetInfo() then
			slave_start()
			mode = nil
		end
	elseif mode == "slave" then
		if ev == "CHAT_MSG_GROUP" then
			if data.msg and data.msg:match("autostack start") then
				launchspeed = tonumber(data.msg:match("autostack start (%d+)")) or 0
				maxspeed = tonumber(data.msg:match("autostack start %d+ (%d+)"))
				start()
				mode = nil
			end
		end
	end
end
RegisterEvent(autostack, "INVENTORY_UPDATE")
RegisterEvent(autostack, "CHAT_MSG_GROUP")


local function autostack_cb(_, args)
	if not args then
		if mode then
			asprint(mode.." mode")
		else
			asprint("off")
		end
	elseif args[1] == "start" then
		launchspeed = tonumber(args[2]) or 0
		maxspeed = tonumber(args[3])

		if mode == "master" then
			slave_start()
		end
		start()
		mode = nil
	elseif args[1] == "stop" then
		stop()
		mode = nil
	elseif args[1] == "master" then
		SendChat("autostack master will relay start commands to slaves...", "GROUP")
		mode = "master"
	elseif args[1] == "slave" then
		SendChat("autostack slave waiting...", "GROUP")
		mode = "slave"
	elseif args[1] == "off" then
		stop()
		mode = nil
	end
end
RegisterUserCommand("autostack", autostack_cb)

